/* $Id: bstream.c,v 1.11 2002/06/14 13:54:04 aa5779 Exp $ */
/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include "util.h"
#include "interp.h"
#define BSTREAM_IMP
#include "bstream.h"
#include "flow.h"
#undef getc
#undef puts
#undef putc

stream_class basic_stream_class = {"text", bs_open, bs_close, bs_read, bs_writestr, 
								   bs_writetag, bs_ioctl,
								   bs_getpos, bs_setpos, bs_rewind
								  };

stream *create_basic(char *name, char *fname, int datasize, int writeable)
{
	stream *ns = ALLOC(stream);
	char *opts = ""; 
	basic_stream *nbs = alloc_chunk(datasize);
	ns->class = &basic_stream_class;
	ns->data = nbs;
	ns->direction = writeable;
	nbs->cooked_mode = 1;
	nbs->size = datasize;
	nbs->fc = detect_flow_class(fname, &fname, &opts);
	nbs->fh = nbs->fc->open(fname, opts, writeable);
	return ns;
}
								   
stream *bs_open(interpreter *i, char *name, char *fname, int writeable)
{
	return create_basic(name, fname, sizeof(basic_stream), writeable);
}

void bs_close(interpreter *i, stream *s)
{
	basic_stream *bsd = s->data;
	bsd->fc->close(bsd->fh);
	free_chunk(bsd, bsd->size);
}

html_chunk *bs_read(interpreter *i, stream *s)
{
	static html_chunk c;
	static html_chunk null_char;
	basic_stream *bs = s->data;
	int ch;

	release_html_chunk(&c);
	ch = bs->fc->getc(bs->fh);
	if(ch == FLOW_EOF)
		return NULL;
	if(ch == FLOW_ERROR)
		recover_error("flow read error: %s", strerror(errno));
	if(ch)
		set_html_char(&c, ch);
	else
	{
		if(HTML_CHUNK_TYPE(&null_char) == hct_none)
			set_tag_name(&null_char, ds_fix(ds_create("NULL")));
		return &null_char;
	}
	return &c;
}

void bs_writestr(interpreter *i, stream *s, char *str)
{
	basic_stream *bs = s->data;
	if(s->direction != stream_out)
		return;
	if(bs->fc->puts(bs->fh, str) == FLOW_ERROR)
		recover_error("flow write error: %s", strerror(errno));
}

void bs_writetag(interpreter *i, stream *s, html_chunk *ch)
{
	basic_stream *bs = s->data;
	if(HTML_CHUNK_TYPE(ch) == hct_char)
	{
		char buf[2] = {0, 0};
		*buf = get_html_char(ch);
		s->class->writestr(i, s, buf);
	}
	else if(HTML_CHUNK_TYPE(ch) == hct_tag)
	{
		if(!strcasecmp(get_tag_name(ch), "NULL"))
		{
			if(bs->fc->putc(bs->fh, 0) == FLOW_ERROR)
				recover_error("flow write error: %s", strerror(errno));
		}
	}
}

char *bs_ioctl(interpreter *i, stream *s, char *ctlstr)
{
	basic_stream *bs = s->data;
	if(!strcasecmp(ctlstr, "providesbasic"))
		return "true";
	else
	if(!strncasecmp(ctlstr, "flow=", 5))
	{
		char *nf = ctlstr + 5;
		char *opts;
		bs->fc->close(bs->fh);
		bs->fc = detect_flow_class(nf, &nf, &opts);
		bs->fh = bs->fc->open(nf, opts, s->direction);
		s->class->rewind(i, s);
		return "";
	}
	/* Attention: a big kludge!!! */
	/*
	else if(!strcasecmp(ctlstr, "__setflow"))
	{
		struct header
		{
			char label[5]; // SETF
			void *fc;
			void *fh;
		} *ptrs = (struct header *)(ctlstr + 10);
		if(strcmp(ptrs->label, "SETF"))
			return NULL;
		bs->fc->close(bs->fh);
		bs->fc  = ptrs->fc;
		bs->fh  = ptrs->fh;
		return "";
	}
	*/
	else if(!strcasecmp(ctlstr, "flush"))
	{
		return bs->fc->flowctl(bs->fh, "flush");
	}
	else if(!strcasecmp(ctlstr, "cooked"))
	{
		bs->cooked_mode++;
		return "";
	}
	else if(!strcasecmp(ctlstr, "raw"))
	{
		if(bs->cooked_mode)
			bs->cooked_mode--;
		return "";
	}
	else if(!strcasecmp(ctlstr, "?cooked"))
		return bs->cooked_mode ? "cooked" : "raw";
	else if(!strcasecmp(ctlstr, "rewind"))
	{
		s->class->rewind(i, s);
		return "";
	}
	else if(!strncasecmp(ctlstr, "flowctl=", 8))
		return bs->fc->flowctl(bs->fh, ctlstr + 8);
	return NULL;
}


int bs_getpos(interpreter *i, stream *s)
{
	basic_stream *bs = s->data;
	return bs->fc->getpos(bs->fh);
}

void bs_setpos(interpreter *i, stream *s, int pos)
{
	basic_stream *bs = s->data;
	if(bs->fc->setpos(bs->fh, pos) == FLOW_ERROR)
		recover_error("can't set flow position: %s", strerror(errno));
}

void bs_rewind(interpreter *i, stream *s)
{
	basic_stream *bs = s->data;
	bs->fc->rewind(bs->fh);
}

stream *dummy_open(interpreter *i, char *name, char *fname, int writeable)
{
	fatal_error("streams: internal error");
	return NULL; // never reached
}

void dummy_close(interpreter *i, stream *s)
{
	// dummy
}

html_chunk *dummy_read(interpreter *i, stream *s)
{
	return NULL;
}

void dummy_writestr(interpreter *i, stream *s, char *str)
{
	// dummy
}

void dummy_writetag(interpreter *i, stream *s, html_chunk *ch)
{
	// dummy
}

char *dummy_ioctl(interpreter *i, stream *s, char *ctlstring)
{
	if(!strcmp(ctlstring, "providesbasic"))
		return "";
	return NULL;
}

int dummy_getpos(interpreter *i, stream *s)
{
	return -1;
}

void dummy_setpos(interpreter *i, stream *s, int pos)
{
	return; // dummy
}

void dummy_rewind(interpreter *i, stream *s)
{
	return; // dummy
}
