#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <math.h>
#include <string.h>
#include <limits.h>
#include <modules.h>
#include <util.h>
#include <lists.h>
#include <dstring.h>
#include <stdlib.h>
#include "version.h"

static module *this_module;
extern procstack *get_pstack(void);
extern variable *safe_get_var(interpreter *i, char *name, int writing);

#define INT_LENGTH 0

static double convert_to_d(char *str)
{
	char *check = "";
	double nv = strtod(str, &check);
	if(*check)
		nv = (double)convert_to_int(str);
	return nv;
}

void fpmath_LTX_init(module *cur_mod)
{
	this_module = cur_mod;
}

static void bi_isfloat(chunk *current, interpreter *i, value *args)
{
	char *str = DS_BODY(args->val);
	char *check = "";
	strtod(str, &check);
	if(!*check)
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_fadd(chunk *current, interpreter *i, value *args)
{
	double value = 0;

	for(; args; args = list_next(args))
		value += convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(value));
}


static void bi_fpostincr(chunk *current, interpreter *i, value *args)
{
	double val = 0;
   	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(get_pstack()->state->buffer, var->vals->val);
	val = convert_to_d(DS_BODY(next_value(args)->val));
	ovalue = convert_to_d(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromdouble(val + ovalue));
}

static void bi_fincr(chunk *current, interpreter *i, value *args)
{
	double val = 0; 
	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	val = convert_to_d(DS_BODY(next_value(args)->val));
	ovalue = convert_to_d(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromdouble(val + ovalue));
	ds_append(get_pstack()->state->buffer, var->vals->val);
}

static void bi_fpostdecr(chunk *current, interpreter *i, value *args)
{
	double val = 0;
   	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(get_pstack()->state->buffer, var->vals->val);
	val = convert_to_d(DS_BODY(next_value(args)->val));
	ovalue = convert_to_d(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromdouble(ovalue - val));
}

static void bi_fdecr(chunk *current, interpreter *i, value *args)
{
	double val = 0.0;
	double ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ovalue = convert_to_d(DS_BODY(var->vals->val));
	val = convert_to_d(DS_BODY(next_value(args)->val));
	ds_assign(var->vals->val, ds_fromdouble(ovalue - val));
	ds_append(get_pstack()->state->buffer, var->vals->val);
}

static void bi_fmul(chunk *current, interpreter *i, value *args)
{
	double value = 1.0;

	for(; args; args = list_next(args))
		value *= convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(value));
}

static void bi_fsub(chunk *current, interpreter *i, value *args)
{
	double v1 = convert_to_d(DS_BODY(args->val));
	double v2 = convert_to_d(DS_BODY(((value	*)list_next(args))->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v1 - v2));
}

static void bi_fdiv(chunk *current, interpreter *i, value *args)
{
	double v1 = convert_to_d(DS_BODY(args->val));
	double v2 = convert_to_d(DS_BODY(((value	*)list_next(args))->val));
	if(!v2)
		recover_error("zero division");
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v1 / v2));
}

static void bi_fneg(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(-v));
}

static void bi_fabs(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	ds_append(get_pstack()->state->buffer, ds_fromdouble(fabs(v)));
}

static void bi_fint(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	modf(v, &v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_frac(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val)), ip;
	v = modf(v, &ip);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_fsqrt(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = sqrt(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_fln(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = log(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_fsin(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = sin(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_fcos(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = cos(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_fexp(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val));
	v = exp(v);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_fpower(chunk *current, interpreter *i, value *args)
{
	double v = convert_to_d(DS_BODY(args->val)), p = convert_to_d(DS_BODY(next_value(args)->val));
	v = pow(v, p);
	ds_append(get_pstack()->state->buffer, ds_fromdouble(v));
}

static void bi_fprecision(chunk *current, interpreter *i, value *args)
{
	int prec;
	if(args)
		prec = ds_dbl_set_precision(convert_to_int(DS_BODY(args->val)));
	else
		prec = ds_dbl_set_precision(-1);
	ds_append(get_pstack()->state->buffer, ds_fromint(prec, 10, 0));
}

static void pass_thru(chunk *current, interpreter *i, value *args, int opcode)
{
	instruction_t *old_vector = get_ma_data(this_module, i);
	if(old_vector)
		old_vector[opcode](current, i, args);
}

enum opcodes_in_question { op_eq = 0, op_neq, op_less, op_greater, op_le, op_ge };
// (refer to the following instruction table)

static void bi_eq(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_eq);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val))  == convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_neq(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_neq);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) != convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_less(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_less);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) < convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_greater(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_greater);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) > convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_le(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_le);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) <= convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static void bi_ge(chunk *current, interpreter *i, value *args)
{
	if(!third_value(args) || strcmp(DS_BODY(args->val), "float"))
		pass_thru(current, i, args, op_ge);
	else
	if(convert_to_d(DS_BODY(next_value(args)->val)) >= convert_to_d(DS_BODY(third_value(args)->val)))
		ds_appendstr(get_pstack()->state->buffer, "true");
}

static master_instruction_src fpmath_add_builtins[] = {
							{"isfloat", 1, bi_isfloat, 1, 1, 1, 1},
							{"fadd", 1, bi_fadd, 2, INT_MAX, 2, 1},
							{"fsub", 1, bi_fsub, 2, 2, 2, 1},
							{"fmul", 1, bi_fmul, 2, INT_MAX, 2, 1},
							{"fdiv", 1, bi_fdiv, 2, 2, 2, 1},
							{"fneg", 1, bi_fneg, 1, 1, 1, 1},
							{"fabs", 1, bi_fabs, 1, 1, 1, 1},
							{"fincr", 1, bi_fincr, 2, 2, 2},
							{"fdecr", 1, bi_fdecr, 2, 2, 2},
							{"fpostincr", 1, bi_fpostincr, 2, 2, 2},
							{"fpostdecr", 1, bi_fpostdecr, 2, 2, 2},
							{"fint", 1, bi_fint, 1, 1, 1, 1},
							{"frac", 1, bi_frac, 1, 1, 1, 1},
							{"fprecision", 1, bi_fprecision, 0, 1, 1},
							{"fln", 1, bi_fln, 1, 1, 1},
							{"fsqrt", 1, bi_fsqrt, 1, 1, 1},
							{"fsin", 1, bi_fsin, 1, 1, 1},
							{"fcos", 1, bi_fcos, 1, 1, 1},
							{"fexp", 1, bi_fexp, 1, 1, 1},
							{"fpower", 1, bi_fpower, 2, 2, 2},
							{NULL}
							};


static master_instruction_src fpmath_builtins[] = {
							{"eq", 1, bi_eq, 2, 3, 2, 1},
							{"neq", 1, bi_neq, 2, 3, 2, 1},
							{"less", 1, bi_less, 2, 3, 2, 1},
							{"greater", 1, bi_greater, 2, 3, 2, 1},
							{"le", 1, bi_le, 2, 3, 2, 1},
							{"ge", 1, bi_ge, 2, 3, 2, 1},
							{NULL}
};

typedef struct fpmath_data
{
	instruction_t old[NUMOFITEMS(fpmath_builtins)];
	int precision;
} fpmath_data;

void fpmath_LTX_attach(mod_attach *ma)
{
	 ma->data = xmalloc(sizeof(fpmath_data));
	 append_builtins(ma->i, fpmath_add_builtins);
	 replace_builtins(ma->i, fpmath_builtins, ((fpmath_data *)ma->data)->old);
}

const char fpmath_LTX_version[] = VERSION;
const char fpmath_LTX_date[] = __DATE__;
